/*
 * Copyright (c) 2011 Miyakawa Taku
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package org.bitbucket.miyakawa_taku.hadoop_job_builder.internal;

import java.lang.reflect.Type;
import java.lang.reflect.ParameterizedType;

import org.apache.hadoop.fs.Path;

/** Common methods. */
public final class Commons {


    /**
     * Returns an array of paths which represents the paths
     * specified by {@code pathNames}.
     *
     * @param pathNames
     *     Path names.
     *
     * @return
     *     Paths representing {@code pathNames}.
     */
    public static Path[] pathsFromPathnames( String[] pathNames ) {
        Path[] paths = new Path[ pathNames.length ];
        for ( int index = 0 ; index < pathNames.length ; ++ index ) {
            paths[ index ] = new Path( pathNames[ index ] );
        }
        return paths;
    }


    /**
     * Returns the class which is top of the given stack trace,
     * excluding {@code exclude}.
     *
     * @param traces
     *     Stack trace elements on which this method search the class.
     *
     * @param exclude
     *     The class specified by this parameter is ignored.
     *
     * @return
     *     The class on the top of the stack trace, except for {@code exclude}.
     *
     * @throws IllegalStateException
     *     If no class is found.
     */
    public static Class< ? > classOnTopOfStack(
            StackTraceElement[] traces , Class< ? > exclude ) {
        for ( StackTraceElement trace : traces ) {
            String className = trace.getClassName();
            if ( className != null && ! className.equals( exclude.getName() ) ) {
                return Commons.findClass( className );
            }
        }
        throw new IllegalStateException( "Cannot find the calling class" );
    }


    /**
     * Finds the class associated with the given {@code className}.
     *
     * <p>If the class is not found, {@linkplain IllegalStateException} is thrown.</p>
     */
    static Class< ? > findClass( String className ) {
        try {
            return Class.forName( className );
        } catch ( ClassNotFoundException exception ) {
            throw new IllegalStateException( "Cannot find " + className , exception );
        }
    }


    /**
     * Checks that {@code type}, which is referred from {@code referer},
     * is a class.
     *
     * @param type
     *     Type to check.
     *
     * @param description
     *     Description of {@code type}.
     *
     * @param referer
     *     Class that refers to {@code type}.
     *
     * @throws IllegalStateException
     *     If {@code type} is not a class.
     */
    public static void checkClass( Type type , String description , Class< ? > referer ) {
        if ( ! ( type instanceof Class ) ) {
            throw new IllegalStateException(
                    "Cannot resolve " + description + " class from " + referer.getName() );
        }
    }


    /**
     * Returns true iff {@code type} is a parameterized type of {@code superType}.
     *
     * @param type
     *     Type to test.
     *
     * @param superType
     *     Supertype to test.
     *
     * @return
     *     {@code true} iff {@code type} is a parameterized type of {@code superType}.
     */
    public static boolean isParameterizedTypeOf( Type type , Class< ? > superType ) {
        return type instanceof ParameterizedType
            && ( (ParameterizedType) type ).getRawType().equals( superType );
    }
}

// vim: et sw=4 sts=4
