/*
 * Copyright (c) 2011 Miyakawa Taku
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package org.bitbucket.miyakawa_taku.hadoop_job_builder.mapreduce;

import org.bitbucket.miyakawa_taku.hadoop_job_builder.internal.Commons;

import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.List;
import java.util.ArrayList;
import java.io.IOException;
import java.net.URI;
import java.net.URISyntaxException;

import org.apache.hadoop.mapreduce.Mapper;
import org.apache.hadoop.mapreduce.Reducer;
import org.apache.hadoop.mapreduce.Partitioner;
import org.apache.hadoop.mapreduce.Job;
import org.apache.hadoop.mapreduce.lib.input.FileInputFormat;
import org.apache.hadoop.mapreduce.lib.input.TextInputFormat;
import org.apache.hadoop.mapreduce.lib.input.SequenceFileInputFormat;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.TextOutputFormat;
import org.apache.hadoop.mapreduce.lib.output.SequenceFileOutputFormat;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.filecache.DistributedCache;
import org.apache.hadoop.conf.Configuration;
import org.apache.hadoop.io.RawComparator;

/**
 * Type-safe builder of {@linkplain Job a job}
 * of {@linkplain org.apache.hadoop.mapreduce the new API of Hadoop MapReduce}.
 *
 * <p>This builder allows comple-time type checking
 * between {@linkplain Job#setMapperClass a mapper},
 * {@linkplain Job#setReducerClass a reducer},
 * {@linkplain Job#setCombinerClass a combiner},
 * {@linkplain Job#setPartitionerClass a partitioner},
 * {@linkplain RawComparator comparators}, keys and values.
 * This also provides convenient methods to set up {@linkplain DistributedCache}
 * and several common {@linkplain Job#setInputFormatClass input formats}
 * and {@linkplain Job#setOutputFormatClass output formats}.</p>
 *
 * <h3>Example</h3>
 *
 * <pre id="simplisticexample">
 * Job job = JobBuilder.of( new WordCountMapper() , new WordCountReducer() )
 *     .jobName( "word-count" )
 *     .detectJar()
 *     .detectKeyValue()
 *     .combiner( new WordCountReducer() )
 *     .inputTextFrom( "word-count/input" )
 *     .outputSequenceFileOn( "word-count/output" )
 *     .buildJob( getConf() );
 * // launches the job
 * job.submit();
 * </pre>
 *
 * <h3>Equivalent code using {@linkplain org.apache.hadoop.mapred the new MapReduce API}</h3>
 *
 * <pre>
 * Job job = new Job( getConf() );
 * job.setJobName( "word-count" );
 * job.setJarByClass( getClass() );
 * job.setMapperClass( WordCountMapper.class );
 * job.setReducerClass( WordCountReducer.class );
 * job.setMapOutputKeyClass( Text.class );
 * job.setMapOutputValueClass( IntWritable.class );
 * job.setOutputKeyClass( Text.class );
 * job.setOutputValueClass( IntWritable.class );
 * job.setCombinerClass( WordCountReducer.class );
 * job.setInputFormatClass( TextInputFormat.class );
 * FileInputFormat.addInputPath( job , new Path( "word-count/input" ) );
 * job.setOutputFormatClass( SequenceFileOutputFormat.class );
 * FileOutputFormat.setOutputPath( job , new Path( "word-count/output" ) );
 * // launches the job
 * job.submit();
 * </pre>
 *
 * <h3>Given these mapper/reducer classes</h3>
 *
 * <pre>
 * public static class WordCountMapper
 *         extends Mapper&lt; LongWritable , Text , Text , IntWritable &gt; {
 *     &#64;Override public void map(...) { ... }
 *     ...
 * }
 *
 * public static class WordCountReducer
 *         extends Reducer&lt; Text , IntWritable , Text , IntWritable &gt; {
 *     &#64;Override public void reduce(...) { ... }
 *     ...
 * }
 * </pre>
 *
 * <h3>Additional configuration</h3>
 *
 * <p>You can of cource set up a job additionally for any purpose.</p>
 *
 * <pre>
 * Job job = JobBuilder.of( new WordCountMapper() , new WordCountReducer() )
 *     .jobName( "word-count" )
 *     .detectJar()
 *     .detectKeyValue()
 *     .combiner( new WordCountReducer() )
 *     .inputTextFrom( "word-count/input" )
 *     .outputSequenceFileOn( "word-count/output" )
 *     .buildJob( getConf() );
 * // additional configuration
 * SequenceFileOutputFormat.setOutputCompressionType( job , BLOCK );
 * // launches the job
 * job.submit();
 * </pre>
 *
 * <h3>Index</h3>
 *
 * <dl>
 *   <dt>Set up the mapper and the reducer</dt>
 *   <dd>{@link #of(Mapper,Reducer)}</dd>
 *   <dd>{@link #of(Mapper)}</dd>
 *
 *   <dt>Set up the keys and values</dt>
 *   <dd>{@link #mapOutputKeyValue(Class,Class)}</dd>
 *   <dd>{@link #outputKeyValue(Class,Class)}</dd>
 *   <dd>{@link #detectKeyValue}</dd>
 *   <dd>{@link #detectMapOutputKeyValue}</dd>
 *   <dd>{@link #detectOutputKeyValue}</dd>
 *
 *   <dt>Set up the combiner</dt>
 *   <dd>{@link #combiner(Reducer)}</dd>
 *
 *   <dt>Set up the partitioner</dt>
 *   <dd>{@link #partitioner(Partitioner)}</dd>
 *
 *   <dt>Set up the comparators</dt>
 *   <dd>{@link #sortComparator(RawComparator)}</dd>
 *   <dd>{@link #groupingComparator(RawComparator)}</dd>
 *
 *   <dt>Set up distributed cache</dt>
 *   <dd>{@link #cacheFile(String)}</dd>
 *   <dd>{@link #cacheFileWithSymlink(String,String)}</dd>
 *   <dd>{@link #cacheArchive(String)}</dd>
 *   <dd>{@link #cacheArchiveWithSymlink(String,String)}</dd>
 *
 *   <dt>Set up the job name</dt>
 *   <dd>{@link #jobName(String)}</dd>
 *
 *   <dt>Set up the jar file name</dt>
 *   <dd>{@link #detectJar}</dd>
 *   <dd>{@link #jarByClass(Class)}</dd>
 *
 *   <dt>Set up the working directory</dt>
 *   <dd>{@link #workingDirectory(Path)}</dd>
 *   <dd>{@link #workingDirectory(String)}</dd>
 *
 *   <dt>Set up the number of reduce tasks</dt>
 *   <dd>{@link #numReduceTasks(int)}</dd>
 *
 *   <dt>Set up the input format</dt>
 *   <dd>{@link #inputTextFrom(Path...)}</dd>
 *   <dd>{@link #inputTextFrom(String...)}</dd>
 *   <dd>{@link #inputSequenceFileFrom(Path...)}</dd>
 *   <dd>{@link #inputSequenceFileFrom(String...)}</dd>
 *   <dd>{@link #inputFileFrom(FileInputFormat,Path...)}</dd>
 *   <dd>{@link #inputFileFrom(FileInputFormat,String...)}</dd>
 *
 *   <dt>Set up the output format</dt>
 *   <dd>{@link #outputTextOn(Path)}</dd>
 *   <dd>{@link #outputTextOn(String)}</dd>
 *   <dd>{@link #outputSequenceFileOn(Path)}</dd>
 *   <dd>{@link #outputSequenceFileOn(String)}</dd>
 *   <dd>{@link #outputFileOn(FileOutputFormat,Path)}</dd>
 *   <dd>{@link #outputFileOn(FileOutputFormat,String)}</dd>
 *
 *   <dt>General purpose</dt>
 *   <dd>{@link #addSetup(JobSetup)}</dd>
 * </dl>
 *
 * @param <KEYIN>
 *     Input key of the job.
 *
 * @param <VALUEIN>
 *     Input value of the job.
 *
 * @param <KEYMID>
 *     Intermediate key of the job.
 *
 * @param <VALUEMID>
 *     Intermediate value of the job.
 *
 * @param <KEYOUT>
 *     Output key of the job.
 *
 * @param <VALUEOUT>
 *     Output value of the job.
 *
 * @since 1.0
 */
public class JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT > {


    /** Mapper class of the job. */
    private final Class< ? extends Mapper > mapperClass;


    /** Reducer class of the job. */
    private final Class< ? extends Reducer > reducerClass;


    /** Actions that set up a job. */
    private final List< JobSetup > setups = new ArrayList< JobSetup >();


    /**
     * Constructs a builder with a mapper and a reducer.
     *
     * <p>This constructor is intended to be accessed from child classes.
     * If you simply want to get a builder instance,
     * use {@linkplain #of(Mapper,Reducer)}.</p>
     *
     * @param mapper
     *     Sample instance of the mapper class of the job.
     *
     * @param reducer
     *     Sample instance of the reducer class of the job.
     *
     * @see #of(Mapper,Reducer)
     *
     * @since 1.0
     */
    protected JobBuilder( Mapper< KEYIN , VALUEIN , KEYMID , VALUEMID > mapper
            , Reducer< KEYMID , VALUEMID , KEYOUT , VALUEOUT > reducer ) {
        this.mapperClass = mapper.getClass();
        this.reducerClass = reducer.getClass();
    }


    /**
     * Constructs a builder with a mapper.
     *
     * <p>This constructor is intended to be accessed from child classes.
     * If you simply want to get a builder instance, use {@linkplain #of(Mapper)}.</p>
     *
     * @param mapper
     *     Sample instance of the mapper class of the job.
     *
     * @see #of(Mapper)
     *
     * @since 1.0
     */
    protected JobBuilder( Mapper< KEYIN , VALUEIN , KEYMID , VALUEMID > mapper ) {
        this.mapperClass = mapper.getClass();
        this.reducerClass = null;
    }


    /**
     * Returns a builder of a job.
     *
     * <h4>Example using generic mapper/reducer classes</h4>
     *
     * <pre>
     * Job job = JobBuilder.of(
     *         new Mapper&lt; Text , Text &gt;()
     *         , new Reducer&lt; Text , Text &gt;() )
     *     .detectJar()
     *     .outputKeyValue( Text.class , Text.class )
     *     .inputKeyValueTextFrom( "input" )
     *     .outputTextOn( "output" )
     *     .buildJob( getConf() );
     * job.submit();
     * </pre>
     *
     * @param mapper
     *     Sample instance of the mapper class of the job.
     *
     * @param reducer
     *     Sample instance of the reducer class of the job.
     *
     * @param <KEYIN>
     *     Input key of the job.
     *
     * @param <VALUEIN>
     *     Input value of the job.
     *
     * @param <KEYMID>
     *     Intermediate key of the job.
     *
     * @param <VALUEMID>
     *     Intermediate value of the job.
     *
     * @param <KEYOUT>
     *     Output key of the job.
     *
     * @param <VALUEOUT>
     *     Output value of the job.
     *
     * @return
     *     Builder of a job.
     *
     * @since 1.0
     */
    public static < KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            of( Mapper< KEYIN , VALUEIN , KEYMID , VALUEMID > mapper
                , Reducer< KEYMID , VALUEMID , KEYOUT , VALUEOUT > reducer ) {
        return new JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            ( mapper , reducer );
    }


    /**
     * Returns a builder of a job with no reducer.
     *
     * <h4>Example using generic a mapper class</h4>
     *
     * <pre>
     * Job job = JobBuilder.of( new Mapper&lt; Text , Text &gt;() )
     *     .detectJar()
     *     .outputKeyValue( Text.class , Text.class )
     *     .inputKeyValueTextFrom( "input" )
     *     .outputTextOn( "output" )
     *     .buildJob( getConf() );
     * job.submit();
     * </pre>
     *
     * @param mapper
     *     Sample instance of the mapper class of the job.
     *
     * @param <KEYIN>
     *     Input key of the job.
     *
     * @param <VALUEIN>
     *     Input value of the job.
     *
     * @param <KEYOUT>
     *     Output key of the job.
     *
     * @param <VALUEOUT>
     *     Output value of the job.
     *
     * @return
     *     Builder of a job.
     *
     * @since 1.0
     */
    public static < KEYIN , VALUEIN , KEYOUT , VALUEOUT >
            JobBuilder< KEYIN , VALUEIN , KEYOUT , VALUEOUT , KEYOUT , VALUEOUT >
            of( Mapper< KEYIN , VALUEIN , KEYOUT , VALUEOUT > mapper ) {
        return new JobBuilder< KEYIN , VALUEIN , KEYOUT , VALUEOUT , KEYOUT , VALUEOUT >
            ( mapper );
    }


    /**
     * Sets the combiner class of the job.
     *
     * @param combiner
     *     Sample instance of the combiner class of the job.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            combiner( Reducer< KEYMID , VALUEMID , KEYMID , VALUEMID > combiner ) {
        final Class< ? extends Reducer > combinerClass = combiner.getClass();
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setCombinerClass( combinerClass );
            }
        } );
        return this;
    }


    /**
     * Sets the partitioner class of the job.
     *
     * @param partitioner
     *     Sample instance of the partitioner class of the job.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            partitioner( Partitioner< KEYMID , VALUEMID > partitioner ) {
        final Class< ? extends Partitioner > partitionerClass = partitioner.getClass();
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setPartitionerClass( partitionerClass );
            }
        } );
        return this;
    }


    /**
     * Sets the sort comparator class of the job.
     *
     * @param sortComparator
     *     Sample instance of the sort comparator class of the job.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            sortComparator( RawComparator< KEYMID > sortComparator ) {
        final Class< ? extends RawComparator > comparatorClass = sortComparator.getClass();
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setSortComparatorClass( comparatorClass );
            }
        } );
        return this;
    }


    /**
     * Sets the grouping comparator class of the job.
     *
     * @param groupingComparator
     *     Sample instance of the grouping comparator class of the job.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            groupingComparator( RawComparator< KEYMID > groupingComparator ) {
        final Class< ? extends RawComparator > comparatorClass = groupingComparator.getClass();
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setGroupingComparatorClass( comparatorClass );
            }
        } );
        return this;
    }


    /**
     * Sets the intermediate key/value classes of the job.
     *
     * @param mapOutputKeyClass
     *     Intermediate key class of the job.
     *
     * @param mapOutputValueClass
     *     Intermediate value class of the job.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            mapOutputKeyValue( final Class< KEYMID > mapOutputKeyClass
                    , final Class< VALUEMID > mapOutputValueClass ) {
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setMapOutputKeyClass( mapOutputKeyClass );
                job.setMapOutputValueClass( mapOutputValueClass );
            }
        } );
        return this;
    }


    /**
     * Sets the output key/value classes of the job.
     *
     * @param outputKeyClass
     *     Output key class of the job.
     *
     * @param outputValueClass
     *     Output value class of the job.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputKeyValue( final Class< KEYOUT > outputKeyClass
                    , final Class< VALUEOUT > outputValueClass ) {
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setOutputKeyClass( outputKeyClass );
                job.setOutputValueClass( outputValueClass );
            }
        } );
        return this;
    }


    /**
     * Detects and sets the key-value classes.
     *
     * <p>Calling this method is equivalent to calling
     * {@linkplain #detectMapOutputKeyValue}
     * and  {@linkplain #detectOutputKeyValue}.</p>
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            detectKeyValue() {
        detectMapOutputKeyValue();
        detectOutputKeyValue();
        return this;
    }


    /**
     * Detects and sets the intermediate key-value classes.
     *
     * <p>This method detects the types from
     * the type parameters of the intermediate key/value pair of the mapper class.
     * The mapper class must inherit {@linkplain Mapper} directly,
     * with raw classes as the type parameters.
     * (It is an implementation limitation. Theoretically, the type inferrence
     * can be much smarter.)</p>
     *
     * <p>If {@linkplain Mapper} is specified as the mapper,
     * this method detects the types from
     * the type parameters of the input key/value pair of the reducer class.</p>
     *
     * @return
     *     {@code this}.
     *
     * @throws IllegalStateException
     *     When the type inferrence fails.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            detectMapOutputKeyValue() {
        KeyValue kv = getMapOutputKeyValue();
        @SuppressWarnings( "unchecked" )
        Class< KEYMID > key = (Class< KEYMID >) kv.key;
        @SuppressWarnings( "unchecked" )
        Class< VALUEMID > value = (Class< VALUEMID >) kv.value;
        mapOutputKeyValue( key , value );
        return this;
    }


    /**
     * Detects and sets the output key-value classes.
     *
     * <p>This method detects the types from
     * the type parameters of the output key/value pair of the reducer class.
     * The reducer class must inherit {@linkplain Reducer} directly,
     * with raw classes as the type parameters.
     * (It is an implementation limitation. Theoretically, the type inferrence
     * can be much smarter.)</p>
     *
     * <p>If no reducer is specified, or {@linkplain Reducer} is specified,
     * this method detects the types from
     * the type parameters of the output key/value pair of the mapper class.</p>
     *
     * @return
     *     {@code this}.
     *
     * @throws IllegalStateException
     *     When the type inferrence fails.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            detectOutputKeyValue() {
        KeyValue kv = getOutputKeyValue();
        @SuppressWarnings( "unchecked" )
        Class< KEYOUT > key = (Class< KEYOUT >) kv.key;
        @SuppressWarnings( "unchecked" )
        Class< VALUEOUT > value = (Class< VALUEOUT >) kv.value;
        outputKeyValue( key , value );
        return this;
    }


    /** Gets intermediate key-value classes from the mapper. */
    private KeyValue getMapOutputKeyValue() {
        if ( this.mapperClass.equals( Mapper.class ) ) {
            return getReduceInputKeyValue();
        }
        ParameterizedType mapperClass
            = getParameterizedSuperclass( this.mapperClass , Mapper.class );
        Type[] typeArgs = mapperClass.getActualTypeArguments();
        Type midKey = typeArgs[ 2 ];
        Type midValue = typeArgs[ 3 ];
        Commons.checkClass( midKey , "key" , this.mapperClass );
        Commons.checkClass( midValue , "value" , this.mapperClass );
        return new KeyValue( (Class) midKey , (Class) midValue );
    }


    /** Gets intermediate key-value classes from the reducer. */
    private KeyValue getReduceInputKeyValue() {
        ParameterizedType reducerSuper
            = getParameterizedSuperclass( this.reducerClass , Reducer.class );
        Type[] typeArgs = reducerSuper.getActualTypeArguments();
        Type midKey = typeArgs[ 0 ];
        Type midValue = typeArgs[ 1 ];
        Commons.checkClass( midKey , "key" , this.reducerClass );
        Commons.checkClass( midValue , "value" , this.reducerClass );
        return new KeyValue( (Class) midKey , (Class) midValue );
    }


    /** Gets output key-value classes. */
    private KeyValue getOutputKeyValue() {
        if ( this.reducerClass == null || this.reducerClass.equals( Reducer.class ) ) {
            return getMapOutputKeyValue();
        }
        ParameterizedType reducerSuper
            = getParameterizedSuperclass( this.reducerClass , Reducer.class );
        Type[] typeArgs = reducerSuper.getActualTypeArguments();
        Type outKey = typeArgs[ 2 ];
        Type outValue = typeArgs[ 3 ];
        Commons.checkClass( outKey , "key" , this.reducerClass );
        Commons.checkClass( outValue , "value" , this.reducerClass );
        return new KeyValue( (Class) outKey , (Class) outValue );
    }


    /**
     * Returns a parameterized superclass of {@code superClass}
     * that is inherited by {@code klass}.
     */
    private static ParameterizedType getParameterizedSuperclass(
            Class < ? > klass , Class< ? > superClass ) {
        Type type = klass.getGenericSuperclass();
        if ( ! Commons.isParameterizedTypeOf( type , superClass ) ) {
            throw new IllegalStateException(
                    "Cannot get generic supertype of " + superClass.getName()
                    + " from " + klass.getName() );
        }
        return (ParameterizedType) type;
    }


    /** Pair of key-class and value-class. */
    static final class KeyValue {

        /** Key-class. */
        final Class< ? > key;

        /** Value-class. */
        final Class< ? > value;

        /** Constructs a pair. */
        KeyValue( Class< ? > key , Class< ? > value ) {
            this.key = key;
            this.value = value;
        }
    }


    /**
     * Adds {@code pathName} as a distributed cache file, using
     * {@linkplain DistributedCache#addCacheFile DistributedCache#addCacheFile}.</p>
     *
     * <p>Your mapper and reducer can read the cache file through
     * {@linkplain DistributedCache#getLocalCacheFiles
     * DistributedCache#getLocalCacheFiles}.</p>
     *
     * <h4>Example</h4>
     *
     * <pre>
     * Job job = JobBuilder.of( new FilterMapper() , new InjectReducer() )
     *     .detectJar()
     *     .detectKeyValue()
     *     .inputKeyValueTextFrom( "input/orders" )
     *     .outputTextOn( "output/statistics" )
     *     .cacheFile( "input/customers.txt" )
     *     .buildJob( getConf() );
     * job.submit();
     * </pre>
     *
     * @param pathName
     *     Path of the cache file.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            cacheFile( String pathName ) {
        final URI cacheURI = uri( pathName );
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                DistributedCache.addCacheFile( cacheURI , job.getConfiguration() );
            }
        } );
        return this;
    }


    /**
     * Adds {@code pathName} as a distributed cache file,
     * and creates a symlink named {@code linkName} in
     * work directories of tasks.
     *
     * <p>Calling this method is equivalent to calling
     * {@code DistributedCache.addCacheFile(
     *     new URI( pathName + "#" + linkName , job.getConfiguration() )}
     * and {@code DistributedCache.createSymlink( job.getConfiguration() )}.</p>
     *
     * <h4>Example</h4>
     *
     * <pre>
     * Job job = JobBuilder.of( new FilterMapper() , new InjectReducer() )
     *     .detectJar()
     *     .detectKeyValue()
     *     .inputKeyValueTextFrom( "input/orders" )
     *     .outputTextOn( "output/statistics" )
     *     .cacheFileWithSymlink( "input/customers.txt" , "customers.txt" )
     *     .buildJob( getConf() );
     * job.submit();
     *
     * public static class InjectReducer extends Reducer&lt;...&gt; {
     *     private List&lt; String &gt; customers;
     *     &#64;Override public void setup( Context context ) {
     *         this.customers = FileUtils.readLines( new File( "customers.txt" ) );
     *     }
     *     &#64;Override public void reduce(...) {
     *         ...
     *     }
     * }
     * </pre>
     *
     * @param pathName
     *     Path of the cache file.
     *
     * @param linkName
     *     Name of the symlink to the cache file.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            cacheFileWithSymlink( String pathName , String linkName ) {
        final URI cacheURI = uri( pathName + "#" + linkName );
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                DistributedCache.addCacheFile( cacheURI , job.getConfiguration() );
                DistributedCache.createSymlink( job.getConfiguration() );
            }
        } );
        return this;
    }


    /**
     * Adds {@code pathName} as a distributed cache archive.
     *
     * @param pathName
     *     Path of the cache archive.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            cacheArchive( String pathName ) {
        final URI cacheURI = uri( pathName );
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                DistributedCache.addCacheArchive( cacheURI , job.getConfiguration() );
            }
        } );
        return this;
    }


    /**
     * Adds {@code pathName} as a distributed cache archive,
     * and creates a symlink named {@code linkName} in
     * work directories of tasktrackers.
     *
     * @param pathName
     *     Path of the cache archive.
     *
     * @param linkName
     *     Name of the symlink to the cache file.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            cacheArchiveWithSymlink( String pathName , String linkName ) {
        final URI cacheURI = uri( pathName + "#" + linkName );
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                DistributedCache.addCacheArchive( cacheURI , job.getConfiguration() );
                DistributedCache.createSymlink( job.getConfiguration() );
            }
        } );
        return this;
    }


    /** Returns a uri that represents {@code pathName}. */
    private static URI uri( String pathName ) {
        try {
            return new URI( pathName );
        } catch ( URISyntaxException e ) {
            throw new IllegalArgumentException( "Illegal URI: " + pathName , e );
        }
    }


    /**
     * Uses {@link TextInputFormat} as the input format of the job
     * and sets {@code paths} as the input files/directories.
     *
     * @param paths
     *     Paths from which the job reads texts.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputTextFrom( Path ... paths ) {
        inputFileFromInternal( TextInputFormat.class , paths );
        return this;
    }


    /**
     * Uses {@link TextInputFormat} as the input format of the job
     * and sets {@code pathNames} as the input files/directories.
     *
     * @param pathNames
     *     Path names from which the job reads texts.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputTextFrom( String ... pathNames ) {
        inputFileFromInternal( TextInputFormat.class , pathNames );
        return this;
    }


    /**
     * Uses {@link SequenceFileInputFormat} as the input format of the job
     * and sets {@code paths} as the input files/directories.
     *
     * @param paths
     *     Paths from which the job reads sequence files.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputSequenceFileFrom( Path ... paths ) {
        inputFileFromInternal( SequenceFileInputFormat.class , paths );
        return this;
    }


    /**
     * Uses {@link SequenceFileInputFormat} as the input format of the job
     * and sets {@code pathNames} as the input files/directories.
     *
     * @param pathNames
     *     Path names from which the job reads sequence files.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputSequenceFileFrom( String ... pathNames ) {
        inputFileFromInternal( SequenceFileInputFormat.class , pathNames );
        return this;
    }


    /**
     * Reads input files from {@code paths} using {@code inputFormat}
     * as the input format of the job.
     *
     * @param inputFormat
     *     Sample instance of the input format of the job,
     *     which extends {@link FileInputFormat}.
     *
     * @param paths
     *     Paths from which the job reads files.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputFileFrom(
                FileInputFormat< KEYIN , VALUEIN > inputFormat
                , Path ... paths ) {
        inputFileFromInternal( inputFormat.getClass() , paths );
        return this;
    }


    /**
     * Reads input files from {@code paths} using {@code inputFormat}
     * as the input format of the job.
     *
     * @param inputFormat
     *     Sample instance of the input format of the job,
     *     which extends {@link FileInputFormat}.
     *
     * @param pathNames
     *     Paths from which the job reads files.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            inputFileFrom(
                    FileInputFormat< KEYIN , VALUEIN > inputFormat
                    , String ... pathNames ) {
        inputFileFromInternal( inputFormat.getClass() , pathNames );
        return this;
    }


    /**
     * Reads input files from {@code paths} using {@code inputFormat}
     * as the input format of the job.
     *
     * @param inputFormat
     *     Class of the input format of the job, which extends {@link FileInputFormat}.
     *
     * @param paths
     *     Paths from which the job reads files.
     */
    private void inputFileFromInternal(
            final Class< ? extends FileInputFormat > inputFormat
            , final Path ... paths ) {
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) throws IOException {
                for ( Path path : paths ) {
                    FileInputFormat.addInputPath( job , path );
                }
                job.setInputFormatClass( inputFormat );
            }
        } );
    }


    /**
     * Reads input files from {@code paths} using {@code inputFormat}
     * as the input format of the job.
     *
     * @param inputFormat
     *     Class of the input format of the job, which extends {@link FileInputFormat}.
     *
     * @param pathNames
     *     Paths from which the job reads files.
     */
    private void inputFileFromInternal(
            Class< ? extends FileInputFormat > inputFormat
            , String ... pathNames ) {
        Path[] paths = Commons.pathsFromPathnames( pathNames );
        inputFileFromInternal( inputFormat , paths );
    }


    /**
     * Uses {@link TextOutputFormat} as the output format
     * and sets {@code path} as the output directory.
     *
     * @param path
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputTextOn( Path path ) {
        outputFileOn( TextOutputFormat.class , path );
        return this;
    }


    /**
     * Uses {@link TextOutputFormat} as the output format
     * and sets {@code pathName} as the output directory.
     *
     * @param pathName
     *     Path name on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputTextOn( String pathName ) {
        return outputTextOn( new Path( pathName ) );
    }


    /**
     * Uses {@link SequenceFileOutputFormat} as the output format
     * and sets {@code path} as the output directory.
     *
     * @param path
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputSequenceFileOn( Path path ) {
        outputFileOn( SequenceFileOutputFormat.class , path );
        return this;
    }


    /**
     * Uses {@link SequenceFileOutputFormat} as the output format
     * and sets {@code pathName} as the output directory.
     *
     * @param pathName
     *     Path name on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            outputSequenceFileOn( String pathName ) {
        return outputSequenceFileOn( new Path( pathName ) );
    }


    /**
     * Outputs the result to the specified output path
     * using the specified output format.
     *
     * @param outputFormat
     *     Sample instance of the output format of the job,
     *     which extends {@linkplain FileOutputFormat}
     *
     * @param path
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT > outputFileOn(
            FileOutputFormat< KEYOUT , VALUEOUT > outputFormat , Path path ) {
        outputFileOn( outputFormat.getClass() , path );
        return this;
    }


    /**
     * Outputs the result to the specified output path
     * using the specified output format.
     *
     * @param outputFormat
     *     Sample instance of the output format of the job,
     *     which extends {@linkplain FileOutputFormat}
     *
     * @param pathName
     *     Path on which the job outputs the result.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT > outputFileOn(
            FileOutputFormat< KEYOUT , VALUEOUT > outputFormat , String pathName ) {
        outputFileOn( outputFormat.getClass() , new Path( pathName ) );
        return this;
    }


    /**
     * Outputs the result to the specified output path
     * using the specified output format.
     *
     * @param outputFormat
     *     Class of the output format of the job,
     *     which extends {@linkplain FileOutputFormat}
     *
     * @param path
     *     Path on which the job outputs the result.
     */
    private void outputFileOn(
            final Class< ? extends FileOutputFormat > outputFormat , final Path path ) {
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                FileOutputFormat.setOutputPath( job , path );
                job.setOutputFormatClass( outputFormat );
            }
        } );
    }


    /**
     * Sets the name of the job.
     *
     * @param jobName
     *     Name of the job.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            jobName( final String jobName ) {
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setJobName( jobName );
            }
        } );
        return this;
    }


    /**
     * Sets the number of reduce tasks.
     *
     * @param numReduceTasks
     *     Number of reduce tasks.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            numReduceTasks( final int numReduceTasks ) {
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setNumReduceTasks( numReduceTasks );
            }
        } );
        return this;
    }


    /**
     * Sets the working directory of the job.
     *
     * @param workingDirectory
     *     Path of the working directory.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            workingDirectory( final Path workingDirectory ) {
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) throws IOException {
                job.setWorkingDirectory( workingDirectory );
            }
        } );
        return this;
    }


    /**
     * Sets the working directory of the job.
     *
     * @param workingDirectoryPathName
     *     Path name of the working directory.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            workingDirectory( final String workingDirectoryPathName ) {
        return workingDirectory( new Path( workingDirectoryPathName ) );
    }


    /**
     * Sets the jar file of the job by an example class.
     *
     * @param exampleClass
     *     Class that represents the jar of the job.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            jarByClass( final Class< ? > exampleClass ) {
        addSetup( new JobSetup< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >() {
            @Override public void setup( Job job ) {
                job.setJarByClass( exampleClass );
            }
        } );
        return this;
    }


    /**
     * Sets the jar file of the job by the class
     * that calls this method.
     *
     * @return
     *     {@code this}.
     *
     * @since 1.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            detectJar() {
        Class< ? > callingClass = Commons.classOnTopOfStack(
                new Throwable().getStackTrace() , getClass() );
        jarByClass( callingClass );
        return this;
    }


    /**
     * Adds a set-up command.
     *
     * @param setup
     *     Set-up command.
     *
     * @return
     *     {@code this}.
     *
     * @see JobSetup
     *     Example
     *
     * @since 2.0
     */
    public JobBuilder< KEYIN , VALUEIN , KEYMID , VALUEMID , KEYOUT , VALUEOUT >
            addSetup( JobSetup< ? super KEYIN , ? super VALUEIN
                    , ? super KEYMID , ? super VALUEMID
                    , ? super KEYOUT , ? super VALUEOUT > setup ) {
        this.setups.add( setup );
        return this;
    }


    /**
     * Builds a job.
     *
     * <p>Calling this method is equivalent to create a new {@link Job job}
     * with the parent configuration {@code baseConf},
     * and call {@link #setup(Job)} with the new job.</p>
     *
     * @param baseConf
     *     Configuration whose settings will be inherited.
     *
     * @return
     *     New job.
     *
     * @throws IOException
     *     When an IO error occurs.
     *
     * @since 1.0
     */
    public Job buildJob( Configuration baseConf ) throws IOException {
        Job job = new Job( baseConf );
        setup( job );
        return job;
    }


    /**
     * Sets up the existing job.
     *
     * <p>The following example code is equivalent
     * to <a href="#simplisticexample">the example code on the top of this page</a>.</p>
     *
     * <pre>
     * Job job = new Job( getConf() );
     * JobBuilder.of( new WordCountMapper() , new WordCountReducer() )
     *     .jobName( "word-count" )
     *     .detectJar()
     *     .detectKeyValue()
     *     .combiner( new WordCountReducer() )
     *     .inputTextFrom( "word-count/input" )
     *     .outputSequenceFileOn( "word-count/output" )
     *     .setup( job );
     * job.submit();
     * </pre>
     *
     * @param job
     *     Job to set up.
     *
     * @throws IOException
     *     When an IO error occurs.
     *
     * @since 1.0
     */
    public void setup( Job job ) throws IOException {
        job.setMapperClass( this.mapperClass );
        if ( this.reducerClass != null ) {
            job.setReducerClass( this.reducerClass );
        }
        for ( JobSetup setup : this.setups ) {
            setup.setup( job );
        }
    }
}

// vim: et sw=4 sts=4
