/*
 * Copyright (c) 2011 Miyakawa Taku
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package org.bitbucket.miyakawa_taku.hadoop_job_builder.mapreduce;

import java.io.IOException;

import org.apache.hadoop.mapreduce.Job;

/**
 * Command to set up a job.
 *
 * <p>Example:</p>
 *
 * <pre>
 * static &lt; KEY , VALUE &gt;
 *         JobSetup&lt; Object , Object , Object , Object , KEY , VALUE &gt;
 *         output( final OutputFormat&lt; KEY , VALUE &gt; outputFormat ) {
 *     return new JobSetup&lt; Object , Object , Object , Object , KEY , VALUE &gt;() {
 *         &#64;Override public setup( Job job ) {
 *             job.setOutputFormatClass( outputFormat.getClass() );
 *         }
 *     };
 * }
 *
 * &#64;Override
 * public int run(String[] args) throws Exception {
 *     Job job = JobBuilder.of( new FooMapper() , new FooReducer() )
 *         .detectJar()
 *         .detectKeyValue()
 *         .inputTextFrom( "/data/foo" )
 *         .addSetup( outputFormat( new NullOutputFormat&lt; Text , Text &gt;() ) )
 *         .buildJob( getConf() );
 *     return ( job.waitForCompletion( true ) ? 0 : 1 );
 * }
 *
 * public static class FooMapper extends Mapper&lt; LongWritable , Text , Text , Text &gt; {
 *     ...
 * }
 *
 * public static class FooReducer extends Reducer&lt; Text , Text , Text , Text &gt; {
 *     ...
 * }
 * </pre>
 *
 * @param <K1>
 *     Super type of the input key.
 *
 * @param <V1>
 *     Super type of the input value.
 *
 * @param <K2>
 *     Super type of the intermediate key.
 *
 * @param <V2>
 *     Super type of the intermediate value.
 *
 * @param <K3>
 *     Super type of the output key.
 *
 * @param <V3>
 *     Super type of the output value.
 *
 * @see JobBuilder#addSetup(JobSetup)
 *
 * @since 2.0
 */
public interface JobSetup< K1 , V1 , K2 , V2 , K3 , V3 > {


    /**
     * Set up a job.
     *
     * @param job
     *     Job to set up.
     *
     * @throws IOException
     *     Thrown when an I/O eror occurs.
     *
     * @since 2.0
     */
    void setup( Job job ) throws IOException;
}

// vim: et sw=4 sts=4
